/* 15-25 November 2023, by Ruud Harmsen
   First completely working version 15 December 2023.

	For midi file specs see https://www.midi.org/specifications/
	but they require a login.

	See also:
	https://ccrma.stanford.edu/~craig/14q/midifile/MidiFileFormat.html
	http://www.music.mcgill.ca/~ich/classes/mumt306/StandardMIDIfileformat.html
	https://www.lim.di.unimi.it/IEEE/MIDI/META.HTM
	http://www.somascape.org/midi/tech/mfile.html
	https://www.euclideanspace.com/art/music/midi/midifile/index.htm
	https://faydoc.tripod.com/formats/mid.htm
	http://www.jososoft.dk/yamaha/articles/midi_10.htm

	See also:
	http://midibox.org/forums/topic/18506-tools-to-debug-midi-files/
		Tools to debug MIDI files?
	https://interglacial.com/pub/dr_midi.html
		Daktari MIDI -- a MIDI file dumper
	https://www.fourmilab.ch/webtools/midicsv/
		Midi to CSV and back
 */

#ifndef RHMIDIFUNCS_H
#define RHMIDIFUNCS_H

#include "rhmusicfuncs.h"
#include <stdio.h>

/***********************************************************************
   Function ReadMidiHeadTrackstarts reads the header of the midifile
   that is opened for reading in fpi. That FILE pointer should be at
   the start of the file when calling the function.

   The three pointer arguments will be used to set the following values:
   - p_NumTracks: The number of Midi tracks in the file.
   - p_WhereNumTracks: The byte position (suitable for fseek) where that
     two byte track counter starts. This makes it easy to update it, e.g.
     for adding a track.
   - p_TicksPerQuarter. The ticks per quarter note, also known as the
     `division' field of the header, if its first bit is zero. If that
     bit is 1, the value the pointer p_TicksPerQuarter points to will
     be left untouched.

   The return value is a calloc'ed array of longs (suitable for fseek),
   containing the file positions where the tracks start. Those positions
   are after the MTrk indication, and after the 4 byte track length
   counter. So it is where the Midi events in the track start.
   The array is calloc'ed, so the calling function should free it when
   being done.

   In case of error, the function returns NULL. If parameter err is NULL,
   the reason for failure is not reported. If it is a valid pointer to
   int, that int will be set to a non-zero error code, as follows:

    1: The Midi file has fewer bytes than needed for a midi header.
    2: The Midi file does not start with the `MThd' midi header marker.
    3: Only the Midi formats 0 and 1 are supported, not 2.
    4: Only ticks per quarter-note supported, not time codes.
    5: Failure to calloc enough space for the start positions of the
       midi tracks.
    6: The function didn't find the start of a Midi track where it was
       expected. You can use ftell on fpi to tell where that was.
 **********************************************************************/
long *ReadMidiHeadTrackstarts (FILE *fpi,
   int *p_NumTracks, long *p_WhereNumTracks,
   int *p_TicksPerQuarter, int *p_err);

/***********************************************************************
   High byte first means big endian encoding. Numbytes is typically
   2 or 4.
   CnvFmHighByteFirst: The number which is encoded in the bytes the buf
      points to, which be put in *p_count.
   CnvToHighByteFirst: The reverse operation.
 **********************************************************************/
void CnvFmHighByteFirst (char *buf, int numbytes, long *p_count);
void CnvToHighByteFirst (char *buf, int numbytes, long count);

/***********************************************************************
   Input is a two byte array, denoting what in a Midi file can follow
   a bend event code, hex En in which n is the Midi channel.
   Returns the corresponding bend in cents, relative to an equidistant
   note.
 **********************************************************************/
float CnvFmMidiBend (char *buf);

/***********************************************************************
   Input is a note bend in cents, relative to an equidistant note,
   between -200 and 200 cents. In the buf that should be at least two
   bytes long, the encoding for Midi will appear, which follows a bend
   event code, hex En in which n is the Midi channel.
 **********************************************************************/
void CnvToMidiBend (char *buf, float cents);

/***********************************************************************
   Input is a one to four byte array, denoting a Delta-T value as it
   appears in Midi files.
   *p_Delta_T is the time difference in  ticks, as a long integer (at
   least 31 bits).
   Returns of bytes used for the encoding.
 **********************************************************************/
int CnvFmVariableLengthValue (unsigned char *buf, long *p_Delta_T);

/***********************************************************************
   Input is a long integer denoting a Midi delta T. In the buf that
   should be at least four bytes long, the encoding for Midi will appear,
   as 1 to four bytes as the need arises. Unneeded bytes are left
   unchanged.
   Returns the number of relevant bytes in the buffer.
 **********************************************************************/
int CnvToVariableLengthValue (unsigned char *buf, long Delta_T);

struct rhmidi_eventdata
{
   int  TrackNumber;  /* Not necessary, but comes handy when testing. */
   long DeltaT;       /* In ticks */
   LLI  SigmaT;       /* In ticks */
   double SigmaQ;     /* In quarternotes */
   int  StatusCodeLastSeen;
   int  FrstCode;     /* Contain FF, F0 or F7, or the Midi event type
                         with the channel cleared. */
   int  ScndCode;     /* Contains the channel in case of a Midi event */
   long TrackFilepos; /* In bytes, from 0 */
   int  EndOfTrack;   /* 0 = normal, 1 = end of track detected */
   long EventDataLen;
   char *EventData;
};

/***********************************************************************
   Function ReadMidiEvent reads a Midi Event from the stream opened in
   fpi. If filepos is non-negative, an fseek do that position is done
   first, otherwise the current position is used.

   Return value:
   Negative indicates end-of-track, event FF-2F seen.
   Zero means success.
   Positive return values indicate various errors:
   1) Premature end of file.
   2) Premature end of file, after Delta-T has been read.
   3) Premature end of file, while reading event data.
   4) Cannot malloc data buffer
 **********************************************************************/
int ReadMidiEvent (
   FILE *fpi, struct rhmidi_eventdata *p_evtdat, int TicksPerQuarter);


/***********************************************************************
 **********************************************************************/
int ReadMidiEventErrHandling (
   FILE *fpi, char *ProgPath, char *MidiFilename,
   struct rhmidi_eventdata *p_evtdat,
   int TicksPerQuarter, int verbose);

#endif
